import json
from urllib.parse import urlparse

from oauthlib.oauth2 import BackendApplicationClient
from requests import Session
from requests.adapters import HTTPAdapter
from requests_oauthlib import OAuth2Session
from six import add_metaclass
from urllib3.util.retry import Retry


class ElectronBondAuthError(Exception):
    pass


class ElectronBondReturnError(Exception):
    pass


def http_meth_factory(meth):
    """Utility method for producing HTTP proxy methods for ElectronBondProxyMethods mixin class.

    Urls are prefixed with the baseurl defined
    """

    def http_method(self, url, *args, **kwargs):
        url = urlparse(url)
        full_url = "/".join([self.config["baseurl"].rstrip("/"),
                             url.path.lstrip("/")])
        result = getattr(self.session, meth)(full_url, *args, **kwargs)
        return result
    return http_method


def retry_session(retries, session=None, backoff_factor=0.3,
                  status_forcelist=(500, 502, 503, 504)):
    session = session or Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


class ElectronBondProxyMethods(type):
    """Metaclass to set up proxy methods for all requests-supported HTTP methods"""
    def __init__(cls, name, parents, dct):

        for meth in ("get", "post", "head", "patch",
                     "put", "delete", "options",):
            fn = http_meth_factory(meth)
            fn.__name__ = meth
            fn.__doc__ = """Proxied :meth:`requests.Session.{}` method from :class:`requests.Session`""".format(
                meth)

            setattr(cls, meth, fn)


@add_metaclass(ElectronBondProxyMethods)
class ElectronBond(object):
    """ElectronBonder Web Client"""

    def __init__(self, **config):
        self.config = {}
        self.config.update(config)

        if not hasattr(self, "session"):
            self.session = retry_session(retries=5)
        self.session.headers.update({"Accept": "application/json",
                                     "User-Agent": "ElectronBond/0.1"})

    def authorize(self):
        """Authorizes the client against the configured microservice instance."""

        resp = self.session.post(
            "/".join([self.config["baseurl"].rstrip("/"), "get-token/"]),
            data={"password": self.config["password"], "username": self.config["username"]})

        if resp.status_code not in [200, 201]:
            raise ElectronBondAuthError(
                "Failed to authorize with status: {}".format(
                    resp.status_code))
        else:
            session_token = json.loads(resp.text)["token"]
            self.session.headers["Authorization"] = "Bearer {}".format(
                session_token)
            return session_token

    def authorize_oauth(self):
        """Authorizes the client using a configured OAuth provider."""
        try:
            client = BackendApplicationClient(
                client_id=self.config["oauth_client_id"])
            oauth = OAuth2Session(client=client)
            token = oauth.fetch_token(
                token_url=f"{self.config['oauth_client_baseurl']}/oauth2/token",
                client_id=self.config["oauth_client_id"],
                client_secret=self.config["oauth_client_secret"])
        except Exception as e:
            raise ElectronBondAuthError(
                f"Failed to authorize OAuth with message: {str(e)}")
        self.session.headers["Authorization"] = f"Bearer {token['access_token']}"
        return token['access_token']

    def get_paged(self, url, *args, **kwargs):
        """get list of json objects from urls of paged items"""
        params = {"page": 1}
        if "params" in kwargs:
            params.update(**kwargs["params"])
            del kwargs["params"]

        current_page = self.get(url, params=params, **kwargs)
        current_json = current_page.json()
        try:
            while len(current_json["results"]) > 0:
                for obj in current_json["results"]:
                    yield obj
                if not current_json.get("next"):
                    break
                params["page"] += 1
                current_page = self.get(url, params=params)
                current_json = current_page.json()
        except Exception:
            raise ElectronBondReturnError(
                "get_paged doesn't know how to handle {}".format(current_json))



import json
import unittest
from unittest.mock import Mock, patch

from electronbonder.client import ElectronBond, ElectronBondAuthError

BASEURL = "http://localhost:8007"
OAUTH_BASEURL = "https://awscognito.com"
OAUTH_CLIENT_ID = "123456789"
OAUTH_CLIENT_SECRET = "987654321"
PATH = "custom/path"


class TestClient(unittest.TestCase):

    def setUp(self):
        self.client = ElectronBond(
            baseurl=BASEURL,
            username="username",
            password="password",
            oauth_client_baseurl=OAUTH_BASEURL,
            oauth_client_id=OAUTH_CLIENT_ID,
            oauth_client_secret=OAUTH_CLIENT_SECRET)

    @patch("requests.Session.post")
    def test_authorize(self, mock_post):
        token = "token_value"
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = json.dumps({"token": token})
        resp = self.client.authorize()
        mock_post.return_value.status_code = 201
        resp = self.client.authorize()
        self.assertEqual(token, resp)
        with self.assertRaises(ElectronBondAuthError):
            mock_post.return_value.status_code = 404
            self.client.authorize()

    @patch("requests_oauthlib.OAuth2Session.fetch_token")
    def test_authorize_oauth(self, mock_token):
        token = "12345"
        mock_token.return_value = {"access_token": token}
        self.client.authorize_oauth()
        self.assertEqual(
            self.client.session.headers["Authorization"],
            f"Bearer {token}")
        mock_token.assert_called_with(
            token_url=f"{OAUTH_BASEURL}/oauth2/token",
            client_id=OAUTH_CLIENT_ID,
            client_secret=OAUTH_CLIENT_SECRET)

    @patch("requests.Session.get")
    def test_get_paged(self, mock_get):
        list(self.client.get_paged(PATH))
        mock_get.assert_called_once()
        mock_get.assert_called_with(
            '{}/{}'.format(BASEURL, PATH), params={'page': 1})
        mock_get.reset_mock()

        list(self.client.get_paged(PATH, params={"page": 4}))
        mock_get.assert_called_once()
        mock_get.assert_called_with(
            '{}/{}'.format(BASEURL, PATH), params={'page': 4})

        expected_results = {"foo": "bar"}
        results_mock = Mock()
        results_mock.side_effect = [
            {"results": [expected_results], "next": True},
            {"results": [expected_results], "next": False}]
        mock_get.return_value.json = results_mock
        results = list(self.client.get_paged(PATH))
        self.assertEqual(results, [expected_results, expected_results])
        mock_get.reset_mock()



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="ElectronBonder",
    url="https://github.com/RockefellerArchiveCenter/ElectronBonder",
    description="Project Electron Client Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Rockefeller Archive Center",
    author_email="archive@rockarch.org",
    version="2.2",
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires=">=3.10",
    install_requires=[
        "requests",
        "requests-oauthlib",
        "six",
    ],
)
